from typing import Any, Literal, Tuple, TypeAlias, Union, overload
from datetime import datetime


SQLObjectNameType: TypeAlias = Union[str, Tuple[str, str], Tuple[str, str, str]]

QuoteMode: TypeAlias = Literal["ansi", "tsql", "postgres"]


def sql_quote_name(inp: SQLObjectNameType, *, mode: QuoteMode = "ansi"):
    if isinstance(inp, str):
        if inp.startswith("#"):  # temp tables names must not be quoted
            assert " " not in inp
            assert "-" not in inp
            assert "'" not in inp
            assert '"' not in inp
            assert "*" not in inp
            assert "/" not in inp
            assert "\\" not in inp
            return inp
        assert "[" not in inp
        assert "]" not in inp
        assert "`" not in inp
        assert '"' not in inp
        if mode == "ansi":
            return '"' + inp + '"'
        if mode == "postgres":
            return "`" + inp + "`"
        return "[" + inp + "]"
    elif len(inp) == 3:
        db = sql_quote_name(inp[0], mode=mode)
        schema_name = sql_quote_name(inp[1], mode=mode)
        tbl_name = sql_quote_name(inp[2], mode=mode)
        return db + "." + schema_name + "." + tbl_name
    else:
        schema_name = sql_quote_name(inp[0], mode=mode)
        tbl_name = sql_quote_name(inp[1], mode=mode)
        return schema_name + "." + tbl_name


def sql_quote_value(vl: Any):
    if vl is None:
        return "null"
    if isinstance(vl, str):
        return "'" + vl.replace("'", "''") + "'"
    if isinstance(vl, float):
        return str(vl)
    if isinstance(vl, int):
        return str(vl)
    if isinstance(vl, bool):
        return "1" if vl else "0"
    if isinstance(vl, datetime):
        return "'" + vl.isoformat() + "'"
    return "'" + str(vl).replace("'", "''") + "'"



from pathlib import Path
from typing import Literal, cast
import fsspec
from .azure_utils import convert_options
import adlfs


class FileSystemDestination:
    def __init__(self, path: str | Path):
        self.path = Path(path)
        self.storage_options = None

    def __truediv__(self, other: str):
        return FileSystemDestination(self.path / other)

    def mkdir(self):
        self.path.mkdir(parents=True, exist_ok=True)

    def get_fs_path(self) -> tuple[fsspec.AbstractFileSystem, str]:
        return (fsspec.filesystem("file"), str(self.path))

    def __str__(self):
        return str(self.path)

    def rm_tree(self):
        if not self.path.exists():
            return
        import shutil

        shutil.rmtree(self.path)

    def exists(self):
        return self.path.exists()

    def upload(self, data: bytes):
        with open(self.path, "wb") as f:
            f.write(data)

    def modified_time(self):
        fs, path = self.get_fs_path()
        return fs.modified(path)

    def remove(self):
        self.path.unlink()

    @property
    def parent(self):
        return FileSystemDestination(self.path.parent)

    def as_path_options(self, flavor: Literal["fsspec", "object_store"]):
        return str(self.path), None

    def as_delta_table(self):
        from deltalake import DeltaTable

        return DeltaTable(self.path)

    def with_suffix(self, suffix: str):
        return FileSystemDestination(self.path.with_suffix(suffix))


class AzureDestination:
    def __init__(self, container: str, path: str, storage_options: dict):
        self.container = container
        self.path = path
        self.storage_options = storage_options

    def to_az_path(self):
        return f"az://{self.container}/{self.path}"

    def mkdir(self):
        pass

    def get_fs_path(self) -> tuple[adlfs.AzureBlobFileSystem, str]:
        opts = cast(dict[str, str], convert_options(self.storage_options, "fsspec"))
        return (fsspec.filesystem("az", **opts), self.to_az_path())

    def upload(self, data: bytes):
        opts = cast(dict[str, str], convert_options(self.storage_options, "fsspec"))

        fs: adlfs.AzureBlobFileSystem = fsspec.filesystem("az", **opts)
        with fs.open(self.to_az_path(), "wb") as f:
            f.write(data)  # type: ignore

    def modified_time(self):
        fs, path = self.get_fs_path()
        return fs.modified(path)

    def __truediv__(self, other: str):
        return AzureDestination(
            self.container,
            self.path.removesuffix("/") + "/" + other,
            self.storage_options,
        )

    @property
    def parent(self):
        return AzureDestination(
            self.container,
            "/".join(self.path.removesuffix("/").split("/")[:-1]),
            self.storage_options,
        )

    def exists(self):
        fs, path = self.get_fs_path()
        return fs.exists(path)

    def rm_tree(self):
        fs, path = self.get_fs_path()
        fs.rm(path, recursive=True)

    def remove(self):
        fs, path = self.get_fs_path()
        fs.rm(path)

    def as_path_options(self, flavor: Literal["fsspec", "object_store"]):
        return self.to_az_path(), cast(
            dict[str, str], convert_options(self.storage_options, flavor)
        )

    def as_delta_table(self):
        from deltalake import DeltaTable

        return DeltaTable(
            self.to_az_path(),
            storage_options=convert_options(self.storage_options, "object_store"),  # type: ignore
        )

    def __str__(self):
        return self.to_az_path()

    def with_suffix(self, suffix: str):
        return AzureDestination(
            self.container, self.path + suffix, self.storage_options
        )


Destination = FileSystemDestination | AzureDestination



from .query import sql_quote_name
from .sql_schema import FieldWithType, _get_col_definition
from typing import TYPE_CHECKING, AsyncIterable
import json

if TYPE_CHECKING:
    import pyodbc
    import pyarrow as pa


async def insert_into_table_via_json(
    *,
    json_batches: AsyncIterable[str],
    table_name: tuple[str, str] | str,
    connection: "pyodbc.Connection",
    schema: list[FieldWithType],
    colnames: list[str] | None = None,
):
    colnames = colnames or [f.name for f in schema]
    cols = ", ".join([sql_quote_name(c) for c in colnames])
    col_defs = ", ".join([_get_col_definition(f, True) for f in schema])
    insert_to_tmp_tbl_stmt = f"INSERT INTO {sql_quote_name(table_name)}({cols}) SELECT {cols} from openjson(?) with ({col_defs})"
    async for batch_json in json_batches:
        with connection.cursor() as cursor:
            cursor.execute(insert_to_tmp_tbl_stmt, (batch_json,))


async def _batch_reader_to_json(reader: "pa.RecordBatchStreamReader"):
    try:
        import polars  # type: ignore

        for batch in reader:  # type: ignore
            pld = polars.from_arrow(batch)  # type: ignore
            assert isinstance(pld, polars.DataFrame)  # type: ignore
            jsond = pld.write_json(row_oriented=True)  # type: ignore
            yield jsond
    except ImportError:
        for batch in reader:
            yield json.dumps(batch.to_pylist())


async def insert_into_table_via_json_from_batches(
    *,
    reader: "pa.RecordBatchReader",
    table_name: tuple[str, str] | str,
    connection: "pyodbc.Connection",
    schema: list[FieldWithType],
    colnames: list[str] | None = None,
):
    colnames = colnames or reader.schema.names or [f.name for f in schema]
    r = _batch_reader_to_json(reader)
    await insert_into_table_via_json(
        json_batches=r,
        schema=schema,
        table_name=table_name,
        colnames=colnames,
        connection=connection,
    )



import pyodbc
import pyarrow as pa
from typing import Union, Tuple

from .sql_schema import FieldWithType
from .query import sql_quote_name

MINIMAL_JSON_COMPATIBILITY_LEVEL = 130


async def pyodbc_insert_into_table(
    *,
    reader: pa.RecordBatchReader,
    table_name: Union[Tuple[str, str], str],
    connection: pyodbc.Connection,
    schema: list[FieldWithType],
    supports_json_insert: bool | None = None,
):
    if supports_json_insert is None:
        from .metadata import get_compatibility_level

        supports_json_insert = (
            get_compatibility_level(connection) >= MINIMAL_JSON_COMPATIBILITY_LEVEL
        )

    if not supports_json_insert:
        colnames = reader.schema.names
        cols = ", ".join([sql_quote_name(c) for c in colnames])
        quots = ", ".join(["?" for _ in colnames])
        insert_to_tmp_tbl_stmt = (
            f"INSERT INTO {sql_quote_name(table_name)}({cols}) VALUES({quots})"
        )

        for batch in reader:  # type: ignore
            with connection.cursor() as cursor:
                lsdt = batch.to_pylist()
                prms = [tuple([item[c] for c in colnames]) for item in lsdt]
                cursor.executemany(insert_to_tmp_tbl_stmt, prms)
    else:
        from .json_insert import insert_into_table_via_json_from_batches

        await insert_into_table_via_json_from_batches(
            reader=reader, table_name=table_name, connection=connection, schema=schema
        )



from datetime import datetime, timezone
from pathlib import Path
from typing import Iterable, Literal, Sequence, TypeVar
from deltalake import DeltaTable, WriterProperties, write_deltalake
from deltalake.exceptions import TableNotFoundError, DeltaError
from arrow_odbc import read_arrow_batches_from_odbc
from deltalake2db.duckdb import apply_storage_options
import asyncio
import sqlglot as sg
from odbc2deltalake.destination import (
    AzureDestination,
    Destination,
    FileSystemDestination,
)
from .sql_schema import get_sql_for_schema
from .query import sql_quote_name
import os
import pyarrow as pa
import pyarrow.dataset as ds
import pyarrow.parquet as pq
import pyodbc
from .metadata import (
    get_primary_keys,
    get_columns,
    table_name_type,
    InformationSchemaColInfo,
)
import json
from deltalake2db import get_sql_for_delta_expr
import duckdb
import time
import sqlglot.expressions as ex
from .sql_glot_utils import table_from_tuple, union
from deltalake2db.sql_utils import read_parquet
from .odbc_utils import build_connection_string
import shutil
import logging

logger = logging.getLogger(__name__)

IS_DELETED_COL_NAME = "__is_deleted"
IS_DELETED_COL_INFO = InformationSchemaColInfo.from_name_type(
    IS_DELETED_COL_NAME, "bit"
)
VALID_FROM_COL_NAME = "__timestamp"
VALID_FROM_COL_INFO = InformationSchemaColInfo.from_name_type(
    VALID_FROM_COL_NAME, "datetimeoffset"
)
IS_FULL_LOAD_COL_NAME = "__is_full_load"
IS_FULL_LOAD_COL_INFO = InformationSchemaColInfo.from_name_type(
    IS_FULL_LOAD_COL_NAME, "bit"
)

T = TypeVar("T")

WRITER_PROPERTIES = WriterProperties(compression="ZSTD")


def _not_none(v: T | None) -> T:
    if v is None:
        raise ValueError("Value is None")
    return v


class DBDeltaPathConfigs:
    DELTA_1_NAME = "delta_1"
    """common data for delta load contains data that has changed after last full load via naive criteria timestamp > last_timestamp (or similar, can also be dates)"""
    DELTA_2_NAME = "delta_2"
    """delta 2 is data where timestamp is different for whatever reason, eg restore"""
    PRIMARY_KEYS_TS = "primary_keys_ts"
    """file with primary keys and timestamps as of now, before load"""
    LAST_PK_VERSION = "last_pk_version"
    """
    file with primary keys and timestamps as of last load
    """

    LATEST_PK_VERSION = "latest_pk_version"
    """file with primary keys and timestamps as of after load. 
      this will be identital to primary_keys file IF there are no updates in the source within the load. 
      this is unlikely, but possible
    """


def _all_nullable(schema: "pa.Schema") -> "pa.Schema":
    import pyarrow.types as pat

    sc = schema
    for i, n in enumerate(schema.names):
        f = schema.field(n)
        if not f.nullable:
            sc = sc.set(i, f.with_nullable(True))
    return sc


def _cast(
    name: str,
    data_type: str,
    *,
    table_alias: str | None = None,
    flavor: Literal["tsql", "duckdb"],
):
    if data_type in ["datetime", "datetime2"] and flavor == "tsql":
        return ex.cast(ex.column(name, table_alias), ex.DataType(this="datetime2(6)"))

    if data_type in ["rowversion", "timestamp"] and flavor == "tsql":
        return ex.cast(ex.column(name, table_alias), ex.DataType.Type.BIGINT)
    return ex.column(name, table_alias)


valid_from_expr = ex.cast(
    ex.func("GETUTCDATE", dialect="tsql"), ex.DataType(this="datetime2(6)")
).as_(VALID_FROM_COL_NAME)


def _get_cols_select(
    cols: list[InformationSchemaColInfo],
    *,
    is_deleted: bool | None = None,
    is_full: bool | None = None,
    with_valid_from: bool = False,
    table_alias: str | None = None,
    flavor: Literal["tsql", "duckdb"],
) -> Sequence[ex.Expression]:
    return (
        [
            _cast(
                c.column_name,
                c.data_type,
                table_alias=table_alias,
                flavor=flavor,
            ).as_(c.column_name)
            for c in cols
        ]
        + ([valid_from_expr] if with_valid_from else [])
        + (
            [ex.cast(ex.convert(int(is_deleted)), "bit").as_(IS_DELETED_COL_NAME)]
            if is_deleted is not None
            else []
        )
        + (
            [ex.cast(ex.convert(int(is_full)), "bit").as_(IS_FULL_LOAD_COL_NAME)]
            if is_full is not None
            else []
        )
    )


def get_delta_col(
    cols: list[InformationSchemaColInfo],
) -> InformationSchemaColInfo | None:
    row_start_col: InformationSchemaColInfo | None = None
    for c in cols:
        if c.data_type.lower() in ["rowversion", "timestamp"]:
            return c
        if c.generated_always_type_desc == "AS_ROW_START":
            row_start_col = c
    return row_start_col


async def write_db_to_delta(
    connection_string: str | dict,
    table: tuple[str, str],
    destination: Destination | Path,
    conn: pyodbc.Connection | None = None,
    *,
    odbc_driver: str | None = None,
):
    if isinstance(destination, Path):
        destination = FileSystemDestination(destination)
    delta_path = destination / "delta"
    owns_con = False
    connection_string = build_connection_string(
        connection_string, odbc=True, odbc_driver=odbc_driver
    )
    if conn is None:
        conn = pyodbc.connect(connection_string)
        owns_con = True
    cols = get_columns(conn, table)[table]

    (destination / "meta").mkdir()
    (destination / "meta/schema.json").upload(
        json.dumps([c.model_dump() for c in cols], indent=4).encode("utf-8")
    )
    if (destination / "delta_load_backup").exists():
        (destination / "delta_load_backup").rm_tree()
    if (destination / "delta_load").exists():
        fs, path = destination.get_fs_path()
        fs.move(
            path + "/" + "delta_load",
            path + "/" + "delta_load_backup",
            recursive=True,
        )
        if (
            destination / f"delta_load_backup/{DBDeltaPathConfigs.LATEST_PK_VERSION}"
        ).exists():
            fs.copy(
                path
                + "/"
                + f"delta_load_backup/{DBDeltaPathConfigs.LATEST_PK_VERSION}",
                path + "/" + f"delta_load/{DBDeltaPathConfigs.LAST_PK_VERSION}",
                recursive=True,
            )
    lock_file_path = destination / "meta/lock.txt"

    try:
        if (
            lock_file_path.exists()
            and (
                datetime.now(tz=timezone.utc) - lock_file_path.modified_time()
            ).total_seconds()
            > 60 * 60
        ):
            lock_file_path.remove()
        lock_file_path.upload(b"")

        delta_col = get_delta_col(cols)  # Use the imported function
        pks = get_primary_keys(conn, table)[table]  # Use the imported functionä

        if not (delta_path / "_delta_log").exists():
            delta_path.mkdir()
            do_full_load(
                connection_string,
                table,
                delta_path,
                mode="overwrite",
                cols=cols,
                pks=pks,
                delta_col=delta_col,
            )
        else:
            if delta_col is None or len(pks) == 0:
                do_full_load(
                    connection_string,
                    table,
                    delta_path,
                    mode="append",
                    cols=cols,
                    pks=pks,
                    delta_col=delta_col,
                )
            else:
                await do_delta_load(
                    connection_string,
                    table,
                    destination=destination,
                    delta_col=delta_col,
                    cols=cols,
                    pks=pks,
                    db_conn=conn,
                )
        lock_file_path.remove()
    except Exception as e:
        # restore files
        if lock_file_path.exists():
            lock_file_path.remove()
        if (destination / "delta_load").exists():
            (destination / "delta_load").rm_tree()
        fs, path = destination.get_fs_path()
        if (destination / "delta_load_backup").exists():
            fs.move(
                path + "/" + "delta_load_backup",
                path + "/" + "delta_load",
                recursive=True,
            )
        raise e
    finally:
        if owns_con:
            conn.close()


def restore_last_pk(
    local_con: duckdb.DuckDBPyConnection,
    destination: Destination,
    delta_table: DeltaTable,
    delta_col: InformationSchemaColInfo,
    pk_cols: list[InformationSchemaColInfo],
):
    delta_table.update_incremental()
    pks = [c.column_name for c in pk_cols]
    with local_con.cursor() as cur:
        sq_valid_from = get_sql_for_delta_expr(
            delta_table,
            select=[ex.func("max", ex.column(VALID_FROM_COL_NAME))],
            conditions=[ex.column(IS_FULL_LOAD_COL_NAME).eq(True)],
        )
        if sq_valid_from is None:
            return False
        cur.execute(sq_valid_from.sql(dialect="duckdb"))
        max_valid_from_res = cur.fetchone()
        latest_full_load_date = max_valid_from_res[0] if max_valid_from_res else None

        last_full_load_expr = get_sql_for_delta_expr(
            delta_table,
            select=_get_cols_select(
                cols=pk_cols + [delta_col],
                table_alias="tr",
                flavor="duckdb",
            ),
            conditions=[
                ex.column(IS_FULL_LOAD_COL_NAME).eq(True),
                ex.column(VALID_FROM_COL_NAME).eq(
                    ex.Subquery(
                        this=ex.select(ex.func("MAX", ex.column(VALID_FROM_COL_NAME)))
                        .from_("tr")
                        .where(ex.column(IS_FULL_LOAD_COL_NAME).eq(True))
                    )
                ),
            ],
            delta_table_cte_name="tr",
        )
        assert last_full_load_expr is not None
        cur.execute(
            "create view last_full_load as " + last_full_load_expr.sql("duckdb")
        )
    with local_con.cursor() as cur:
        sq = get_sql_for_delta_expr(
            delta_table,
            select=_get_cols_select(
                cols=pk_cols + [delta_col, IS_DELETED_COL_INFO],
                table_alias="tr",
                flavor="duckdb",
            ),
            delta_table_cte_name="tr",
            conditions=[
                ex.column(VALID_FROM_COL_NAME) > ex.convert(latest_full_load_date)
            ],
        )
        assert sq is not None
        sq = sq.qualify(
            ex.EQ(
                this=ex.Window(
                    this=ex.RowNumber(),
                    partition_by=[ex.column(pk) for pk in pks],
                    order=ex.Order(
                        expressions=[
                            ex.Ordered(
                                this=ex.column(delta_col.column_name),
                                desc=True,
                                nulls_first=False,
                            )
                        ]
                    ),
                    over="OVER",
                ),
                expression=ex.convert(1),
            )
        )
        cur.execute("create view delta_after_full_load as " + sq.sql("duckdb"))
    with local_con.cursor() as cur:

        sql = f"""create view v_last_pk_version as
                with base as ( 
                select df.* from delta_after_full_load df
                union all
                select f.*, False as {IS_DELETED_COL_NAME} from last_full_load f 
                    anti join delta_after_full_load d on {' AND '.join([f"f.{sql_quote_name(c.column_name)} = d.{sql_quote_name(c.column_name)}" for c in pk_cols])}
                )
                select * EXCLUDE ({IS_DELETED_COL_NAME}) from base where {IS_DELETED_COL_NAME}
                    """
        cur.execute(sql)

    with local_con.cursor() as cur:
        cur.execute("select count(*) from v_last_pk_version")
        res = cur.fetchone()
        assert res is not None
        res = res[0]
        if res == 0:
            return False
    with local_con.cursor() as cur:
        cur.execute("select * from v_last_pk_version")
        rdr = cur.fetch_record_batch()
        dp, do = (
            destination / f"delta_load/{DBDeltaPathConfigs.LAST_PK_VERSION}"
        ).as_path_options(flavor="object_store")
        write_deltalake(
            dp,
            rdr,
            mode="overwrite",
            schema_mode="overwrite",
            writer_properties=WRITER_PROPERTIES,
            engine="rust",
            storage_options=do,
        )
        return True


def create_replace_view(
    conn: duckdb.DuckDBPyConnection,
    name: str,
    format: Literal["parquet", "delta"],
    base_destination: Destination,
):
    with conn.cursor() as cur:
        if format == "parquet":
            location = base_destination / f"delta_load/{name}.parquet"
            cur.execute(
                f"create or replace view {name} as select * from read_parquet('{str(location)}')"
            )
        else:
            location = base_destination / f"delta_load/{name}"
            cur.execute(
                f"create or replace view {name} as "
                + _not_none(get_sql_for_delta_expr(location.as_delta_table())).sql(
                    "duckdb"
                )
            )


def write_latest_pk(
    destination: Destination,
    pks: list[InformationSchemaColInfo],
    delta_col: InformationSchemaColInfo,
):
    with duckdb.connect() as local_con:
        fs_opts = destination.as_path_options("fsspec")[1]
        if fs_opts:
            apply_storage_options(local_con, fs_opts)
        delta_1_path = (
            destination / f"delta_load/{DBDeltaPathConfigs.DELTA_1_NAME}.parquet"
        ).as_path_options("fsspec")[0]
        delta_2_path = (
            destination / f"delta_load/{DBDeltaPathConfigs.DELTA_2_NAME}.parquet"
        )  # we don't always have this one
        current_pk_path = (
            destination / f"delta_load/{DBDeltaPathConfigs.PRIMARY_KEYS_TS}.parquet"
        ).as_path_options("fsspec")[0]

        with local_con.cursor() as cur:
            select_expr = ex.select(
                *_get_cols_select(
                    cols=pks + [delta_col],
                    table_alias="t",
                    flavor="duckdb",
                )
            )
            if delta_2_path.exists():
                create_replace_view(
                    local_con, DBDeltaPathConfigs.DELTA_2_NAME, "parquet", destination
                )
            else:
                cur.execute(
                    "create view delta_2 as "
                    + select_expr.from_(read_parquet(delta_1_path).as_("t"))
                    .where("1=0")
                    .sql("duckdb")
                )

        with local_con.cursor() as cur:
            latest_pk_query = union(
                [
                    ex.select(
                        *_get_cols_select(
                            cols=pks + [delta_col],
                            table_alias="au",
                            flavor="duckdb",
                        )
                    ).from_(table_from_tuple("delta_2", alias="au")),
                    (
                        ex.select(
                            *_get_cols_select(
                                cols=pks + [delta_col],
                                table_alias="d1",
                                flavor="duckdb",
                            )
                        )
                        .from_(read_parquet(delta_1_path).as_("d1"))
                        .join(
                            ex.table_("delta_2", alias="au2"),
                            ex.and_(
                                *[
                                    ex.column(c.column_name, "d1").eq(
                                        ex.column(c.column_name, "au2")
                                    )
                                    for c in pks
                                ]
                            ),
                            join_type="anti",
                        )
                    ),
                    (
                        ex.select(
                            *_get_cols_select(
                                cols=pks + [delta_col],
                                table_alias="cpk",
                                flavor="duckdb",
                            )
                        )
                        .from_(read_parquet(current_pk_path).as_("cpk"))
                        .join(
                            ex.table_("delta_2", alias="au3"),
                            ex.and_(
                                *[
                                    ex.column(c.column_name, "cpk").eq(
                                        ex.column(c.column_name, "au3")
                                    )
                                    for c in pks
                                ]
                            ),
                            join_type="anti",
                        )
                        .join(
                            read_parquet(delta_1_path).as_("au4"),
                            ex.and_(
                                *[
                                    ex.column(c.column_name, "cpk").eq(
                                        ex.column(c.column_name, "au4")
                                    )
                                    for c in pks
                                ]
                            ),
                            join_type="anti",
                        )
                    ),
                ],
                distinct=False,
            )
            cur.execute(latest_pk_query.sql("duckdb"))
            dp, do = (
                destination / f"delta_load/{DBDeltaPathConfigs.LATEST_PK_VERSION}"
            ).as_path_options(flavor="object_store")
            write_deltalake(
                dp,
                cur.fetch_record_batch(),
                mode="overwrite",
                schema_mode="overwrite",
                writer_properties=WRITER_PROPERTIES,
                engine="rust",
                storage_options=do,
            )


async def do_delta_load(
    connection_string: str,
    table: table_name_type,
    destination: Destination,
    *,
    delta_col: InformationSchemaColInfo,
    cols: list[InformationSchemaColInfo],
    pks: list[str],
    db_conn: pyodbc.Connection,
):
    last_pk_path = destination / f"delta_load/{DBDeltaPathConfigs.LAST_PK_VERSION}"
    logger.info(
        f"{table}: Start Delta Load with Delta Column {delta_col.column_name} and pks: {', '.join(pks)}"
    )

    with duckdb.connect() as local_con:
        fs_opts = destination.as_path_options("fsspec")[1]
        if fs_opts:
            apply_storage_options(local_con, fs_opts)

        if not (last_pk_path / "_delta_log").exists():  # or do a full load?
            logger.warning(f"{table}: Primary keys missing, try to restore")
            if not restore_last_pk(
                local_con,
                destination,
                (destination / "delta").as_delta_table(),
                delta_col,
                [c for c in cols if c.column_name in pks],
            ):
                logger.warning(f"{table}: No primary keys found, do a full load")
                do_full_load(
                    connection_string,
                    table,
                    destination / "delta",
                    mode="append",
                    cols=cols,
                    pks=pks,
                    delta_col=delta_col,
                )
                return
        delta_path = destination / "delta"
        dt = delta_path.as_delta_table()
        sq = get_sql_for_delta_expr(
            dt,
            select=[
                ex.func(
                    "MAX",
                    _cast(
                        delta_col.column_name,
                        delta_col.data_type,
                        flavor="duckdb",
                    ),
                )
            ],
        )
        assert sq is not None
        sql = sq.sql("duckdb")
        with local_con.cursor() as cur:
            cur.execute(sql)
            res = cur.fetchone()
            delta_load_value = res[0] if res else None
        if delta_load_value is None:
            logger.warning(f"{table}: No delta load value, do a full load")
            do_full_load(
                connection_string,
                table,
                delta_path,
                mode="append",
                cols=cols,
                pks=pks,
                delta_col=delta_col,
            )
            return
        pk_cols = [c for c in cols if c.column_name in pks]
        pk_ds_cols = pk_cols + [delta_col]
        assert len(pk_ds_cols) == len(pks) + 1
        logger.info(
            f"{table}: Start delta step 1, get primary keys and timestamps. MAX({delta_col.column_name}): {delta_load_value}"
        )
        _retrieve_primary_key_data(
            connection_string=connection_string,
            table=table,
            delta_col=delta_col,
            pk_cols=pk_cols,
            destination=destination,
        )

        criterion = _cast(
            delta_col.column_name, delta_col.data_type, table_alias="t", flavor="tsql"
        ) > ex.convert(delta_load_value)
        logger.info(f"{table}: Start delta step 2, load updates by timestamp")
        _load_updates_to_delta(
            connection_string,
            table,
            delta_path,
            cols=cols,
            local_con=local_con,
            criterion=criterion,
            delta_name="delta_1",
        )
        dt.update_incremental()

        await _handle_additional_updates(
            connection_string=connection_string,
            table=table,
            delta_path=delta_path,
            db_conn=db_conn,
            local_con=local_con,
            pk_cols=pk_cols,
            delta_col=delta_col,
            cols=cols,
        )
        dt.update_incremental()

        logger.info(f"{table}: Start delta step 3.5, write meta for next delta load")

        write_latest_pk(destination, pk_cols, delta_col)

        logger.info(f"{table}: Start delta step 4.5, write deletes")
        do_deletes(
            destination=destination,
            local_con=local_con,
            delta_table=dt,
            cols=cols,
            pk_cols=pk_cols,
        )
        logger.info(f"{table}: Done delta load")


def do_deletes(
    destination: Destination,
    local_con: duckdb.DuckDBPyConnection,
    delta_table: DeltaTable,
    cols: list[InformationSchemaColInfo],
    pk_cols: list[InformationSchemaColInfo],
):
    create_replace_view(
        local_con, DBDeltaPathConfigs.LATEST_PK_VERSION, "delta", destination
    )
    create_replace_view(
        local_con, DBDeltaPathConfigs.LAST_PK_VERSION, "delta", destination
    )
    delete_query = ex.except_(
        left=ex.select(
            *_get_cols_select(pk_cols, table_alias="lpk", flavor="duckdb")
        ).from_(table_from_tuple(DBDeltaPathConfigs.LAST_PK_VERSION, alias="lpk")),
        right=ex.select(
            *_get_cols_select(pk_cols, table_alias="cpk", flavor="duckdb")
        ).from_(table_from_tuple(DBDeltaPathConfigs.LATEST_PK_VERSION, alias="cpk")),
    )

    with local_con.cursor() as cur:
        cur.execute("create view deletes as " + delete_query.sql("duckdb"))
    with local_con.cursor() as cur:
        non_pk_cols = [c for c in cols if c not in pk_cols]
        non_pk_select = [ex.Null().as_(c.column_name) for c in non_pk_cols]
        deletes_with_schema = union(
            [
                ex.select(
                    *_get_cols_select(
                        pk_cols,
                        table_alias="d1",
                        flavor="duckdb",
                    )
                )
                .select(
                    *_get_cols_select(
                        non_pk_cols,
                        table_alias="d1",
                        flavor="duckdb",
                    ),
                    append=True,
                )
                .select(
                    ex.AtTimeZone(
                        this=ex.CurrentTimestamp(),
                        zone=ex.Literal(this="UTC", is_string=True),
                    ).as_(VALID_FROM_COL_NAME),
                    ex.convert(True).as_(IS_DELETED_COL_NAME),
                    ex.convert(False).as_(IS_FULL_LOAD_COL_NAME),
                )
                .from_(table_from_tuple("delta_1", alias="d1"))
                .where("1=0"),  # only used to get correct datatypes
                ex.select(
                    ex.Column(
                        this=ex.Star(), table=ex.Identifier(this="d", quoted=False)
                    )
                )
                .select(*non_pk_select, append=True)
                .select(
                    ex.AtTimeZone(
                        this=ex.CurrentTimestamp(),
                        zone=ex.Literal(this="UTC", is_string=True),
                    ).as_(VALID_FROM_COL_NAME),
                    append=True,
                )
                .select(ex.convert(True).as_(IS_DELETED_COL_NAME), append=True)
                .select(ex.convert(False).as_(IS_FULL_LOAD_COL_NAME), append=True)
                .from_(table_from_tuple("deletes", alias="d")),
            ],
            distinct=False,
        )

        cur.execute(
            "create view deletes_with_schema as " + deletes_with_schema.sql("duckdb")
        )
    with local_con.cursor() as cur:
        cur.execute(
            "select count(*) from (select * from deletes_with_schema limit 1) s"
        )
        res = cur.fetchone()
        assert res is not None
        res = res[0]
        has_deletes = res > 0
    if has_deletes:
        with local_con.cursor() as cur:
            cur.execute(deletes_with_schema.sql("duckdb"))
            rdr = cur.fetch_record_batch()
            write_deltalake(
                delta_table,
                rdr,
                schema=_all_nullable(rdr.schema),
                mode="append",
                schema_mode="merge",
                writer_properties=WRITER_PROPERTIES,
                engine="rust",
            )


def _retrieve_primary_key_data(
    connection_string: str,
    table: table_name_type,
    delta_col: InformationSchemaColInfo,
    pk_cols: list[InformationSchemaColInfo],
    destination: Destination,
):
    pk_ts_col_select = ex.select(
        *_get_cols_select(
            is_full=None,
            is_deleted=None,
            cols=pk_cols + [delta_col],
            with_valid_from=False,
            flavor="tsql",
        )
    ).from_(table_from_tuple(table))
    pk_ts_reader_sql = pk_ts_col_select.sql("tsql")

    pk_path = destination / f"delta_load/{DBDeltaPathConfigs.PRIMARY_KEYS_TS}.parquet"

    write_sql_to_parquet(
        sql=pk_ts_reader_sql, connection_string=connection_string, parquet_path=pk_path
    )
    return pk_path

    # todo: test
    # todo: deletes
    # todo: persist latest ts per pk?


async def _handle_additional_updates(
    connection_string: str,
    table: table_name_type,
    delta_path: Destination,
    db_conn: pyodbc.Connection,
    local_con: duckdb.DuckDBPyConnection,
    pk_cols: list[InformationSchemaColInfo],
    delta_col: InformationSchemaColInfo,
    cols: list[InformationSchemaColInfo],
):
    """Handles updates that are not logical by their timestamp. This can happen on a restore from backup, for example."""
    folder = delta_path.parent
    pk_ds_cols = pk_cols + [delta_col]
    create_replace_view(
        local_con, DBDeltaPathConfigs.PRIMARY_KEYS_TS, "parquet", folder
    )
    create_replace_view(local_con, DBDeltaPathConfigs.LAST_PK_VERSION, "delta", folder)
    with local_con.cursor() as cur:
        cur.execute(
            ex.except_(
                left=ex.select(
                    *_get_cols_select(
                        cols=pk_ds_cols,
                        table_alias="pk",
                        flavor="duckdb",
                    )
                ).from_(
                    table_from_tuple(DBDeltaPathConfigs.PRIMARY_KEYS_TS, alias="pk")
                ),
                right=ex.select(
                    *_get_cols_select(
                        cols=pk_ds_cols,
                        table_alias="lpk",
                        flavor="duckdb",
                    )
                ).from_(
                    table_from_tuple(DBDeltaPathConfigs.LAST_PK_VERSION, alias="lpk")
                ),
            ).sql("duckdb")
        )
        rdr = cur.fetch_record_batch()
        additional_updates_path = folder / "delta_load/additional_updates.parquet"
        write_batch_to_parquet(additional_updates_path, rdr)
    with local_con.cursor() as cur:
        sql_query = ex.except_(
            left=ex.select(
                *_get_cols_select(
                    cols=pk_cols,
                    table_alias="au",
                    flavor="duckdb",
                )
            ).from_(read_parquet(str(additional_updates_path)).as_("au")),
            right=ex.select(
                *_get_cols_select(
                    cols=pk_cols,
                    table_alias="d1",
                    flavor="duckdb",
                )
            ).from_(table_from_tuple("delta_1", alias="d1")),
        )
        cur.execute(
            """create or replace view real_additional_updates as """
            + sql_query.sql("duckdb")
        )
    with local_con.cursor() as cur:
        cur.execute(
            "select count(*) from (select * from real_additional_updates limit 1) s"
        )
        res = cur.fetchone()
        assert res is not None
        res = res[0]
        has_additional_updates = res > 0
    if has_additional_updates:
        logger.warning(f"{table}: Start delta step 3, load strange updates")
        temp_table_name = "##temp_updates_" + str(hash(table[1]))[1:]
        sql = get_sql_for_schema(
            temp_table_name,
            [p.as_field_type() for p in pk_cols],
            primary_keys=None,
            with_exist_check=False,
        )
        with db_conn.cursor() as cur:
            cur.execute(sql)
        from .odbc_insert import pyodbc_insert_into_table

        await pyodbc_insert_into_table(
            reader=local_con.execute(
                "select * from real_additional_updates"
            ).fetch_record_batch(),
            table_name=temp_table_name,
            connection=db_conn,
            schema=[p.as_field_type() for p in pk_cols],
        )

        def _collate(c: InformationSchemaColInfo):
            if c.data_type.lower() in [
                "char",
                "varchar",
                "nchar",
                "nvarchar",
                "text",
                "ntext",
            ]:
                return "COLLATE DATABASE_DEFAULT"
            return ""

        criterion = f"""
            inner join {temp_table_name} ttt on {' AND '.join([f't.{sql_quote_name(c.column_name)} {_collate(c)} = ttt.{sql_quote_name(c.column_name)}' for c in pk_cols])}"""

        _load_updates_to_delta(
            connection_string=connection_string,
            table=table,
            delta_path=delta_path,
            cols=cols,
            local_con=local_con,
            criterion=criterion,
            delta_name="delta_2",
        )
    else:
        if (folder / "delta_load/delta_2.parquet").exists():
            (folder / "delta_load/delta_2.parquet").remove()


def _load_updates_to_delta(
    connection_string: str,
    table: table_name_type,
    delta_path: Destination,
    cols: list[InformationSchemaColInfo],
    local_con: duckdb.DuckDBPyConnection,
    criterion: str | Sequence[str | ex.Expression] | ex.Expression,
    delta_name: str,
):
    if isinstance(criterion, ex.Expression):
        criterion = [criterion]
    delta_sql = (
        ex.select(
            *_get_cols_select(
                cols,
                is_full=False,
                is_deleted=False,
                with_valid_from=True,
                table_alias="t",
                flavor="tsql",
            )
        )
        .where(*(criterion if not isinstance(criterion, str) else []), dialect="tsql")
        .from_(table_from_tuple(table, alias="t"))
        .sql("tsql")
    )
    if isinstance(criterion, str):
        delta_sql += " " + criterion

    delta_name_path = delta_path.parent / f"delta_load/{delta_name}.parquet"
    write_sql_to_parquet(
        sql=delta_sql, connection_string=connection_string, parquet_path=delta_name_path
    )
    with local_con.cursor() as cur:
        cur.execute(
            f'create view {delta_name} as select * from read_parquet("{str(delta_name_path)}")'
        )

    with local_con.cursor() as cur:
        cur.execute(f"select count(*) from (select * from {delta_name} limit 1)")
        res = cur.fetchone()
        assert res is not None
        res = res[0]
        if res == 0:
            return
    with local_con.cursor() as cur:
        cur.execute(f"select * from {delta_name}")
        rdr = cur.fetch_record_batch()
        dp, do = delta_path.as_path_options(flavor="object_store")
        write_deltalake(
            dp,
            rdr,
            storage_options=do,
            schema=_all_nullable(rdr.schema),
            mode="append",
            writer_properties=WRITER_PROPERTIES,
            engine="rust",
            schema_mode="merge",
        )


def write_sql_to_parquet(connection_string: str, sql: str, parquet_path: Destination):
    logger.debug(f"SQL for {parquet_path}: \n{sql}")
    batch_reader = read_arrow_batches_from_odbc(
        query=sql,
        connection_string=connection_string,
        max_binary_size=20000,
        max_text_size=20000,
    )
    temp_path = parquet_path.with_suffix(".tmp.parquet")
    if temp_path.exists():
        temp_path.remove()
    fs, path = temp_path.get_fs_path()
    with pq.ParquetWriter(path, filesystem=fs, schema=batch_reader.schema) as writer:
        for batch in batch_reader:
            writer.write_batch(batch)
    if parquet_path.exists():
        parquet_path.remove()
    fs.move(path, str(parquet_path))


def write_batch_to_parquet(
    parquet_path: Destination, batch_reader: pa.RecordBatchReader
):
    temp_path = parquet_path.with_suffix(".tmp.parquet")
    if temp_path.exists():
        temp_path.remove()
    fs, path = temp_path.get_fs_path()
    with pq.ParquetWriter(path, filesystem=fs, schema=batch_reader.schema) as writer:
        for batch in batch_reader:
            writer.write_batch(batch)
    if parquet_path.exists():
        parquet_path.remove()
    fs.move(path, str(parquet_path))


def do_full_load(
    connection_string: str,
    table: table_name_type,
    delta_path: Destination,
    mode: Literal["overwrite", "append"],
    cols: list[InformationSchemaColInfo],
    delta_col: InformationSchemaColInfo | None,
    pks: list[str],
):
    logger.info(f"{table}: Start Full Load")
    sql = (
        ex.select(
            *_get_cols_select(
                is_deleted=False,
                is_full=True,
                cols=cols,
                with_valid_from=True,
                flavor="tsql",
            )
        )
        .from_(table_from_tuple(table))
        .sql("tsql")
    )
    reader = read_arrow_batches_from_odbc(
        query=sql,
        connection_string=connection_string,
        max_binary_size=20000,
        max_text_size=20000,
    )
    try:
        dt = delta_path.as_delta_table()
        old_add_actions = dt.get_add_actions().to_pylist()
        old_paths = [ac["path"] for ac in old_add_actions]
    except TableNotFoundError:
        old_paths = []
        dt = None
    dp, do = delta_path.as_path_options(flavor="object_store")
    try:
        write_deltalake(
            dp,
            reader,
            schema=_all_nullable(reader.schema),
            mode=mode,
            writer_properties=WRITER_PROPERTIES,
            schema_mode="overwrite" if mode == "overwrite" else "merge",
            engine="rust",
            storage_options=do,
        )
    except DeltaError as e:
        if "No data source supplied to write command" in str(e):
            if mode == "overwrite":
                write_empty_delta_table(reader.schema, dp, do)
            logger.warning(f"{table}: No data to write")
        else:
            raise e

    if delta_col is None:
        logger.info(f"{table}: Full Load done")
        return
    logger.info(f"{table}: Full Load done, write meta for delta load")

    dt = dt or delta_path.as_delta_table()
    dt.update_incremental()
    (delta_path.parent / "delta_load").mkdir()
    with duckdb.connect() as local_con:
        if delta_path.storage_options is not None:
            apply_storage_options(local_con, delta_path.storage_options)
        sql = get_sql_for_delta_expr(
            dt,
            select=[pk for pk in pks] + ([delta_col.column_name] if delta_col else []),
            action_filter=lambda ac: ac["path"] not in old_paths,
        )
        if sql is None:
            return
        with local_con.cursor() as cur:
            cur.execute(sql.sql("duckdb"))
            dp, do = (
                delta_path.parent / f"delta_load/{DBDeltaPathConfigs.LATEST_PK_VERSION}"
            ).as_path_options("object_store")
            write_deltalake(
                dp,
                cur.fetch_record_batch(),
                mode="overwrite",
                schema_mode="overwrite",
                writer_properties=WRITER_PROPERTIES,
                engine="rust",
                storage_options=do,
            )


def write_empty_delta_table(
    schema: pa.Schema, path: str | Path, storage_options: dict[str, str] | None
):
    write_deltalake(
        path,
        [],
        schema=_all_nullable(schema),
        mode="overwrite",
        schema_mode="overwrite",
        engine="pyarrow",
        storage_options=storage_options,
    )



from typing import Optional, Union, Any
from .query import sql_quote_value, sql_quote_name
from odbc2deltalake.metadata import FieldWithType

table_name_type = Union[str, tuple[str, str]]


def _get_sql_type(type_str: str, max_str_length: int | None) -> str:
    if type_str in ["binary", "varbinary", "large_binary"]:
        real_max_length = max_str_length or 8000
        if real_max_length > 8000:
            real_max_length = "MAX"
        return f"varbinary({real_max_length})"
    if type_str in [
        "varchar",
        "nvarchar",
        "string",
        "large_string",
        "utf8",
        "large_utf8",
    ]:
        real_max_length = max_str_length or 4000
        if real_max_length <= 2000:
            real_max_length = real_max_length * 2
        elif real_max_length > 4000:  # need max type
            real_max_length = "MAX"
        return f"nvarchar({real_max_length})"
    return type_str


def sql_quote_value_with_type(
    type_str: str, max_str_length: int | None, value: Any
) -> str:
    if value is not None and type_str.lower() in ["bit", "bool", "boolean"]:
        assert type(value) == bool
        if value == True:
            return f"CAST(1 as bit)"
        else:
            return f"CAST(0 as bit)"
    if value is not None and type_str.lower() in ["string", "int32", "varchar", "int"]:
        return sql_quote_value(value)  # string / int32 are defaults for sql server
    sql_type = _get_sql_type(type_str, max_str_length)
    if value is not None:
        return f"CAST({sql_quote_value(value)} as {sql_type})"
    else:
        return f"CAST(NULL AS {sql_type})"


def _get_col_definition(
    field: FieldWithType,
    nullable: bool,
) -> str:
    sql_type = _get_sql_type(field.type, field.max_str_length)
    definit = (
        sql_quote_name(field.name)
        + " "
        + sql_type
        + (" NOT NULL" if not nullable else "")
    )
    return definit


def get_sql_for_schema(
    table_name: table_name_type,
    schema: list[FieldWithType],
    primary_keys: list[str] | None,
    with_exist_check: bool,
):
    cols_sql = [
        _get_col_definition(
            f,
            primary_keys is None or f.name not in primary_keys,
        )
        for f in schema
    ]

    cols = ", ".join(cols_sql)
    pkdef = ""
    if primary_keys and len(primary_keys) > 0:
        pkcols = ", ".join((sql_quote_name(n) for n in primary_keys))
        tbl_name_pk = (
            table_name.removeprefix("##")
            if isinstance(table_name, str)
            else table_name[0] + "_" + table_name[1]
        )
        pkdef = (
            f", CONSTRAINT {sql_quote_name('PK_'+tbl_name_pk)}  PRIMARY KEY({pkcols})"
        )
    create_sql = f"CREATE TABLE {sql_quote_name(table_name)}({cols}{pkdef}) "
    if with_exist_check:
        return f"""
            IF OBJECT_ID (N'{sql_quote_name(table_name).replace("'", "''")}', N'U') IS NULL 
            BEGIN
                {create_sql}
                select 'created' as action
            END
            else
            begin
                select 'nothing' as action
            end 
        """
    return create_sql



from dataclasses import dataclass
from typing import Callable, Literal, Union, Optional, Any, TYPE_CHECKING
import logging

from pydantic import BaseModel


table_name_type = Union[str, tuple[str, str]]

if TYPE_CHECKING:
    import pyodbc


def get_primary_keys(
    conn: "pyodbc.Connection", table_name: table_name_type | None
) -> dict[table_name_type, list[str]]:
    with conn.cursor() as cur:
        cur.execute(
            """            
    exec sp_executesql N' SELECT ccu.TABLE_SCHEMA, ccu.TABLE_NAME, ccu.COLUMN_NAME
    FROM INFORMATION_SCHEMA.TABLE_CONSTRAINTS tc WITH(NOLOCK)
        JOIN INFORMATION_SCHEMA.CONSTRAINT_COLUMN_USAGE ccu WITH(NOLOCK) ON tc.CONSTRAINT_NAME = ccu.Constraint_name
    WHERE tc.CONSTRAINT_TYPE = ''Primary Key'' AND (ccu.Table_Name=@Table or @Table is null) AND (ccu.TABLE_SCHEMA=@Schema or @Schema is null) ', N'@Table varchar(100), @Schema varchar(100)',
		@Schema=?, @Table=?
""",
            (
                table_name[0] if table_name else None,
                table_name[1] if table_name else None,
            ),
        )
        res = cur.fetchall()
        if res is None:
            return {}
        res_dict = {}
        for r in res:
            tn = r[0], r[1]
            if tn not in res_dict:
                res_dict[tn] = []
            res_dict[tn].append(r[2])
        if table_name and table_name not in res_dict:
            res_dict[table_name] = []
        return res_dict


class FieldWithType(BaseModel):
    name: str
    type: str
    max_str_length: int | None = None


class InformationSchemaColInfo(BaseModel):
    column_name: str
    data_type: str
    column_default: str | None = None
    is_nullable: bool = True
    character_maximum_length: int | None = None
    numeric_precision: int | None = None
    numeric_scale: int | None = None
    datetime_precision: int | None = None
    generated_always_type_desc: Literal[
        "NOT_APPLICABLE", "AS_ROW_START", "AS_ROW_END"
    ] = "NOT_APPLICABLE"

    def as_field_type(self):

        return FieldWithType(
            name=self.column_name,
            type=self.data_type,
            max_str_length=self.character_maximum_length,
        )

    @staticmethod
    def from_name_type(name: str, type: str) -> "InformationSchemaColInfo":
        if "(" in type:
            type, rest = type.split("(", 1)
            rest = rest[:-1]
            if "," in rest:
                precision, scale = rest.split(",", 1)
                precision = int(precision)
                scale = int(scale)
            else:
                precision = int(rest)
                scale = None
        else:
            precision = None
            scale = None
        return InformationSchemaColInfo(
            column_name=name,
            column_default=None,
            is_nullable=True,
            data_type=type,
            character_maximum_length=(
                precision if type in ["varchar", "char", "nvarchar", "nchar"] else None
            ),
            numeric_precision=precision if type in ["decimal", "numeric"] else None,
            numeric_scale=scale if type in ["decimal", "numeric"] else None,
            datetime_precision=None,
            generated_always_type_desc="NOT_APPLICABLE",
        )


def get_columns(
    conn: "pyodbc.Connection", table_name: table_name_type | None
) -> dict[table_name_type, list[InformationSchemaColInfo]]:
    with conn.cursor() as cur:
        cur.execute(
            """            
     exec sp_executesql N' SELECT  ccu.TABLE_SCHEMA, ccu.TABLE_NAME,
		ccu.column_name, ccu.column_default,
		cast(case when ccu.IS_NULLABLE=''YES'' THEN 1 ELSE 0 end as bit) as is_nullable,
		data_type,
		character_maximum_length,
		numeric_precision,
		numeric_scale,
		datetime_precision,
        ci.generated_always_type_desc FROM INFORMATION_SCHEMA.COLUMNS ccu
        left join (
			
SELECT sc.name as schema_name, t.name as table_name, c.name as col_name, c.generated_always_type_desc FROM sys.columns c 
	inner join sys.tables t on t.object_id=c.object_id
	inner join sys.schemas sc on sc.schema_id=t.schema_id
		) ci on ci.schema_name=ccu.TABLE_SCHEMA and ci.table_name=ccu.TABLE_NAME and ci.col_name=ccu.COLUMN_NAME
		where (ccu.Table_Name=@Table or @Table is null) AND (ccu.TABLE_SCHEMA=@Schema or @Schema is null) ', N'@Table varchar(100), @Schema varchar(100)',
		@Schema=?, @Table=?
""",
            (
                table_name[0] if table_name else None,
                table_name[1] if table_name else None,
            ),
        )
        res = cur.fetchall()
        if res is None:
            return {}
        res_dict = {}
        assert cur.description is not None
        col_names = [d[0] for d in cur.description]
        for r in res:
            r_dict = dict(zip(col_names, r))
            tn = r_dict.pop("TABLE_SCHEMA"), r_dict.pop("TABLE_NAME")
            if tn not in res_dict:
                res_dict[tn] = []
            res_dict[tn].append(InformationSchemaColInfo(**r_dict))
        return res_dict


def get_compatibility_level(connection: "pyodbc.Connection") -> int:
    with connection.cursor() as cur:
        cur.execute(
            """
    SELECT compatibility_level  
    FROM sys.databases WHERE name =DB_NAME()"""
        )
        res = cur.fetchone()
        assert res is not None
        return res[0]
